"""Contains the DataAccess class, and functions for working with database transactions."""
import logging
from contextlib import contextmanager

from bidon.util import get_value


class DataAccess(object):
  """A thin wrapper over a DB API2 database connection.

  Provides a context manager, direct access to the execute method, and wrappers over some common
  methods to reduce the amount of boilerplate SQL required.
  """
  def __init__(self, core, *, logger=None):
    """Initialize a DataAccess object with a DataAccessCore"""
    self.core = core
    self.connection = None
    self.last_query = None
    self.row_count = -1
    self._logger = logger

  def __enter__(self):
    """Opens DataAccess and returns self as context manager."""
    return self.open()

  def __exit__(self, ex_type, ex_value, ex_traceback):
    """Closes DataAccess, requesting commit if no execption was thrown."""
    self.close(ex_type == None)

  def _log(self, message, loglevel=logging.INFO):
    """Logs message if logger is present."""
    if self._logger:
      self._logger.log(loglevel, message)

  @property
  def sql_writer(self):
    """The SQL writer object associated with the core."""
    return self.core.sql_writer

  def commit(self):
    """Commit the pending statements."""
    self._log("Commit", logging.DEBUG)
    self.connection.commit()

  def rollback(self):
    """Rollback the pending statements."""
    self._log("Rollback", logging.DEBUG)
    self.connection.rollback()

  @property
  def autocommit(self):
    """Get the autocommit value."""
    return self.core.get_autocommit(self.connection)

  @autocommit.setter
  def autocommit(self, value):
    """Set the autocommit value."""
    self._log("Setting autocommit from {} to {}".format(self.autocommit, value), logging.DEBUG)
    self.core.set_autocommit(self.connection, value)

  def open(self, autocommit=False):
    """Sets the connection via the core's open method."""
    if self.connection is not None:
      raise Exception("Connection already set")
    self.connection = self.core.open()
    self.autocommit = autocommit
    return self

  def close(self, commit=True):
    """Closes the connection via the core's close method."""
    self.core.close(self.connection, commit)
    self.connection = None
    return self

  def execute(self, query_string, params=None):
    """Execute a given query."""
    cr = self.connection.cursor()
    self._log("SQL: {0} ({1})".format(query_string, params))
    self.last_query = (query_string, params)
    cr.execute(query_string, params or self.core.empty_params)
    self._update_cursor_stats(cr)
    return cr

  def callproc(self, proc_name, params):
    """Call a stored procedure."""
    cr = self.connection.cursor()
    self._log(
      "SQL: {0}({1})".format(proc_name, ",".join([str(p) for p in params]) if params else ""))
    self.last_query = (proc_name, params)
    result_params = cr.callproc(proc_name, params)
    self._update_cursor_stats(cr)
    return (cr, result_params)

  def get_callproc_signature(self, proc_name, param_types):
    """Gets a procedure's signature from the name and list of types."""
    if isinstance(param_types[0], (list, tuple)):
      params = [self.core.sql_writer.to_placeholder(*pt) for pt in param_types]
    else:
      params = [self.core.sql_writer.to_placeholder(None, pt) for pt in param_types]

    return proc_name + self.core.sql_writer.to_tuple(params)

  def find(self, table_name, constraints=None, *, columns=None, order_by=None):
    """Find the first record that matches the given criteria.

    Constraints is any construct that canb e parsed by SQLWriter.parse_constraints.
    """
    query_string, params = self.core.sql_writer.get_find_all_query(
      table_name, constraints, columns=columns, order_by=order_by)
    query_string += " limit 1;"
    return self.execute(query_string, params).fetchone()

  def find_all(self, table_name, constraints=None, *, columns=None, order_by=None, limiting=None):
    """Find all records that match a given criteria.

    Constraints is any construct that canb e parsed by SQLWriter.parse_constraints.
    """
    query_string, params = self.core.sql_writer.get_find_all_query(
      table_name, constraints, columns=columns, order_by=order_by, limiting=limiting)
    query_string += ";"
    return self.execute(query_string, params)

  def page(self, table_name, paging, constraints=None, *, columns=None, order_by=None,
           get_count=True):
    """Performs a find_all method with paging. Returns results and the total number of records.

    Paging is a tuple containing (page, page_size). Page is 0-baesd."""
    if get_count:
      count = self.count(table_name, constraints)
    else:
      count = None

    page, page_size = paging

    limiting = None
    if page_size > 0:
      limiting = (page_size, page * page_size)

    records = list(self.find_all(
      table_name, constraints, columns=columns, order_by=order_by, limiting=limiting))
    return (records, count)

  def update(self, table_name, values, constraints=None, *, returning=None):
    """Builds and executes and update statement.

    Values can be either a dict or an enuerable of 2-tuples in the form (column, value). Constraints
    can be any construct that can be parsed by SQLWriter.parse_constraints. However, you cannot mix
    tuples and dicts between values and constraints.
    """
    if constraints is None:
      constraints = "1=1"
    assignments, assignment_params = self.core.sql_writer.parse_constraints(
      values, ", ", is_assignment=True)
    where, where_params = self.core.sql_writer.parse_constraints(constraints, " and ")
    returns = ""
    if returning and self.core.supports_returning_syntax:
      returns = " returning {0}".format(returning)
    sql = "update {0} set {1} where {2}{3};".format(table_name, assignments, where, returns)
    params = assignment_params

    if constraints is None or isinstance(constraints, str):
      pass
    elif isinstance(constraints, dict):
      if isinstance(params, list):
        raise ValueError("you cannot mix enumerable and dict values and constraints")
      params = params or {}
      params.update(where_params)
    else:
      if isinstance(params, dict):
        raise ValueError("you cannot mix enumerable and dict values and constraints")
      params = params or []
      params.extend(where_params)

    cr = self.execute(sql, params)
    return cr

  def insert(self, table_name, values, *, returning=None):
    """Builds and executes an insert statement.

    Values can be either a dict or an enumerable of 2-tuples in the form (column, value).
    """
    if isinstance(values, dict):
      names = values.keys()
      placeholders = [self.core.sql_writer.to_placeholder(i) for i in names]
      params = values
    else:
      names = [i[0] for i in values]
      placeholders = [self.core.sql_writer.to_placeholder() for i in values]
      params = [i[1] for i in values]
    placeholders = self.core.sql_writer.to_tuple(placeholders)
    names = self.core.sql_writer.to_tuple(names)
    returns = ""
    if returning and self.core.supports_returning_syntax:
      returns = " returning {0}".format(returning)
    sql = "insert into {0} {1} values {2}{3};".format(table_name, names, placeholders, returns)
    cr = self.execute(sql, params)
    return cr

  def delete(self, table_name, constraints=None):
    """Builds and executes an delete statement.

    Constraints can be any construct that can be parsed by SQLWriter.parse_constraints.
    """
    if constraints is None:
      constraints = "1=1"
    where, params = self.core.sql_writer.parse_constraints(constraints)
    sql = "delete from {0} where {1};".format(table_name, where)
    self.execute(sql, params)

  def count(self, table_name, constraints=None, *, extract="index"):
    """Get the count of records in a table.

    If the default cursor is a tuple or named tuple, this method will work without specifying an
    extract parameter. If it is a dict cursor, it is necessary to specify any value other than
    'index' for extract. This method will not work with cursors that aren't like tuple, namedtuple
    or dict cursors.
    """
    where, params = self.core.sql_writer.parse_constraints(constraints)
    sql = "select count(*) as count from {0} where {1};".format(table_name, where or "1 = 1")
    # NOTE: Won't work right with dict cursor
    return self.get_scalar(self.execute(sql, params), 0 if extract == "index" else "count")

  def get_scalar(self, cursor, index=0):
    """Get a single value from the first returned record from a cursor.

    By default it will get cursor.fecthone()[0] which works with tuples and namedtuples. For dict
    cursor it is necessary to specify index. This method will not work with cursors that aren't
    indexable.
    """
    if isinstance(index, int):
      return cursor.fetchone()[index]
    else:
      return get_value(cursor.fetchone(), index)

  def _update_cursor_stats(self, cursor):
    """Sets DataAccess members according to the given cursor."""
    self.row_count = cursor.rowcount


@contextmanager
def transaction(dax):
  """Wrap statements in a transaction. If the statements succeed, commit, otherwise rollback."""
  old_autocommit = dax.autocommit
  dax.autocommit = False
  try:
    yield dax
  except Exception as ex:
    dax.rollback()
    raise ex
  else:
    dax.commit()
  finally:
    dax.autocommit = old_autocommit


@contextmanager
def autocommit(dax):
  """Make statements autocommit."""
  if not dax.autocommit:
    dax.commit()
  old_autocommit = dax.autocommit
  dax.autocommit = True
  try:
    yield dax
  finally:
    dax.autocommit = old_autocommit



"""A collection of converters, and converter creators

For the purposes of this module, a converter is any function that takes a single value and returns
another value.
"""
from decimal import Decimal
from uuid import UUID, uuid4
from datetime import datetime, timezone

try:
  import arrow
except ImportError:
  pass


def identity(val):
  """Returns the value it is given."""
  return val


def to_int(val):
  """Convert val to an int."""
  if isinstance(val, int):
    return val
  return int(val)


def to_float(val):
  """Convert val to a float."""
  if isinstance(val, float):
    return val
  return float(val)


def to_decimal(val):
  """Convert val to a decimal."""
  if isinstance(val, Decimal):
    return val
  return Decimal(val)


def to_bit(val):
  """Convert val to a bit: either 1 if val evaluates to True, or False otherwise."""
  return 1 if val else 0


def to_bool(val):
  """Convert val to a boolean value."""
  return bool(val)


def to_uuid(val):
  """Convert val to a uuid."""
  if isinstance(val, UUID):
    return val
  return UUID(val)


def to_compressed_string(val, max_length=0):
  """Converts val to a compressed string.

  A compressed string is one with no leading or trailing spaces.

  If val is None, or is blank (all spaces) None is returned.

  If max_length > 0 and  the stripped val is greater than max_length, val[:max_length] is returned.
  """
  if val is None or len(val) == 0:
    return None
  rval = " ".join(val.split())
  if len(rval) == 0:
    return None
  if max_length == 0:
    return rval
  else:
    return rval[:max_length]


def to_date(val, fmt=None):
  """Convert the value to a date, using arrow."""
  return _to_datetime(val, fmt).date()


def to_datetime(val, fmt=None):
  """Convert the value to a datetime, using arrow."""
  return _to_datetime(val, fmt).datetime


def to_time(val, fmt="HH:mm:ss"):
  """Convert the value to a time, using arrow."""
  return _to_datetime(val, fmt).time()


def _to_datetime(val, fmt=None):
  """Convert the value to a datetime, using arrow."""
  if val is None:
    raise Exception("Invalid date subject")
  if fmt is None or not isinstance(val, str):
    return arrow.get(val)
  else:
    return arrow.get(val, fmt)


def to_now(_):
  """Returns the current timestamp, with a utc timezone."""
  return datetime.utcnow().replace(tzinfo=timezone.utc)


def to_none(_):
  """Returns None."""
  return None


def to_true(_):
  """Returns True."""
  return True


def to_false(_):
  """Returns False."""
  return False


def to_empty_string(_):
  """Returns the empty string."""
  return ""


def to_new_uuid(_):
  """Returns a new randomly generated UUID."""
  return uuid4()


def incrementor(start=0, step=1):
  """Returns a function that first returns the start value, and returns previous value + step on
  each subsequent call.
  """
  def fxn(_):
    """Returns the next value in the sequnce defined by [start::step)"""
    nonlocal start
    rval = start
    start += step
    return rval
  return fxn


def string_trimmer(max_length=0):
  """The same as partial(to_compressed_string(max_length=max_length))"""
  return lambda val: to_compressed_string(val, max_length)


def static_value(val):
  """Returns a function that always returns val."""
  return lambda _: val


def formatted_datetime(fmt):
  """The same as partial(to_datetime(fmt=fmt))."""
  return lambda val: _to_datetime(val, fmt).datetime


def rounded_decimal(places):
  """Returns a lambda that converts its value to a decimal and then rounds the value by places."""
  return lambda val: round(to_decimal(val), places)


def index_resolver(index, strict=False):
  """Returns a function that accepts a value and returns index[value]."""
  if strict:
    return lambda id_: index[id_]
  else:
    return index.get


def accept_none_wrapper(fxn):
  """Wraps a function, returning None if None is passed in, and otherwise passing the
  value along to the given function.
  """
  def wrapper(val):
    """If val is None, return None, otherwise pass the value along to fxn."""
    if val is None:
      return None
    else:
      return fxn(val)
  return wrapper


def try_wrapper(fxn):
  """Wraps a function, returning (True, fxn(val)) if successful, (False, val) if not."""
  def wrapper(val):
    """Try to call fxn with the given value. If successful, return (True, fxn(val)), otherwise
    returns (False, val).
    """
    try:
      return (True, fxn(val))
    except Exception:
      return (False, val)
  return wrapper



"""Contains ModelAccess class and helper methods."""
from .data_access import DataAccess
from bidon.util import local_now, utc_now


class ModelAccess(DataAccess):
  """A model-based specialization of DataAccess

  Provides methods for CRUDing objects derived from bidonß.data.ModelBase.
  """

  def _find_model(self, constructor, table_name, constraints, *, columns=None, order_by=None):
    """Calls DataAccess find and passes the results to the given constructor."""
    data = self.find(table_name, constraints, columns=columns, order_by=order_by)
    return constructor(data) if data else None

  def _find_all_models(self, constructor, table_name, constraints=None, *, columns=None,
                       order_by=None, limiting=None):
    """Calls DataAccess find and passes the results to the given constructor."""
    for record in self.find_all(table_name, constraints, columns=columns, order_by=order_by,
                                limiting=limiting):
      yield constructor(record)

  def find_model(self, constructor, constraints, *, columns=None, table_name=None, order_by=None):
    """Specialization of DataAccess find that returns a model instead of cursor object."""
    return self._find_model(constructor, table_name or constructor.table_name, constraints,
                            columns=columns, order_by=order_by)

  def find_all_models(self, constructor, constraints=None, *, columns=None, order_by=None,
                      limiting=None, table_name=None):
    """Specialization of DataAccess find_all that returns models instead of cursor objects."""
    return self._find_all_models(
      constructor, table_name or constructor.table_name, constraints, columns=columns,
      order_by=order_by, limiting=limiting)

  def page_models(self, constructor, paging, constraints=None, *, columns=None, order_by=None):
    """Specialization of DataAccess page that returns models instead of cursor objects."""
    records, count = self.page(constructor.table_name, paging, constraints, columns=columns,
                               order_by=order_by)
    return ([constructor(r) for r in records], count)

  def find_model_by_id(self, constructor, id_, *, columns=None):
    """Searches for a model by id, according to its class' primary_key_name.

    If primary_key_name is a tuple, id_ must be a tuple with a matching length.
    """
    return self.find_model(
      constructor, get_id_constraints(constructor.primary_key_name, id_), columns=columns)

  def refresh_model(self, model, overwrite=False):
    """Pulls the model's record from the database. If overwrite is True, the model values are
    overwritten and returns the model, otherwise a new model instance with the newer record is
    returned.
    """
    new_model = self.find_model_by_id(model.__class__, model.primary_key)
    if overwrite:
      model.update(new_model.to_dict(use_default_excludes=False))
      return model
    else:
      return new_model

  def update_model(self, model, *, include_keys=None):
    """Updates a model.

    If include_keys is specified, only updates the given attributes. Otherwise, updates all non-id
    attributes.
    """
    id_constraints = get_model_id_constraints(model)

    if include_keys == None:
      include_keys = set(
        model.attrs.keys()).difference(model.exclude_keys_sql).difference(id_constraints.keys())

    # If include_keys was not null but was empty
    if not include_keys:
      return model

    values = model.to_dict(include_keys=include_keys)
    returnings = []

    _, updated_ts = model.timestamps if model.timestamps else (None, None)

    if updated_ts and updated_ts not in values:
      values[updated_ts] = utc_now() if self.core.supports_timezones else local_now()
      returnings.append(updated_ts)

    returning = ", ".join(returnings)
    cr = self.update(model.table_name, values, id_constraints, returning=returning)

    if returning and self.core.supports_returning_syntax:
      rec = cr.fetchone()
      for idx, attr_name in enumerate(returnings):
        setattr(model, attr_name, rec[idx])

    return model

  def insert_model(self, model):
    """Inserts a record for the given model.

    If model's primary key is auto, the primary key will be set appropriately.
    """
    pkname = model.primary_key_name
    include_keys = set(model.attrs.keys()).difference(model.exclude_keys_sql)

    if model.primary_key_is_auto and pkname in include_keys:
      include_keys.remove(pkname)

    data = model.to_dict(include_keys=include_keys)
    returnings = []

    if model.primary_key_is_auto:
      returnings.append(pkname)
    if model.timestamps:
      returnings.extend(ts_name for ts_name in model.timestamps if ts_name)

    returning = ", ".join(returnings)
    cr = self.insert(model.table_name, data, returning=returning)

    if self.core.supports_returning_syntax:
      if returning:
        rec = cr.fetchone()
        for idx, attr_name in enumerate(returnings):
          setattr(model, attr_name, rec[idx])
    else:
      if model.primary_key_is_auto:
        setattr(model, model.primary_key_name, cr.lastrowid)

    return model

  def delete_model(self, model_or_type, id_=None):
    """Deletes a model.

    If model_or_type is a model, delete that model. If it is a ModelBase subclass, id_ must be
    specified, and the associated record is deleted.
    """
    if not id_:
      constraints = get_model_id_constraints(model_or_type)
    else:
      constraints = get_id_constraints(model_or_type.primary_key_name, id_)
    self.delete(model_or_type.table_name, constraints)

    return model_or_type

  def find_or_build(self, constructor, props):
    """Looks for a model that matches the given dictionary constraints. If it is not found, a new
    model of the given type is constructed and returned.
    """
    model = self.find_model(constructor, props)
    return model or constructor(**props)

  def find_or_create(self, constructor, props):
    """Looks for a model taht matches the given dictionary constraints. If it is not found, a new
    model of the given type is created and saved to the database, then returned.
    """
    model = self.find_model(constructor, props)
    if model is None:
      model = constructor(**props)
      self.insert_model(model)
    return model


def get_model_id_constraints(model):
  """Returns constraints to target a specific model."""
  pkname = model.primary_key_name
  pkey = model.primary_key
  return get_id_constraints(pkname, pkey)


def get_id_constraints(pkname, pkey):
  """Returns primary key consraints.

  If pkname is a string, returns a dict with pkname=pkey. pkname and pkey must be enumerables of
  matching length.
  """
  if isinstance(pkname, str):
    return {pkname: pkey}
  else:
    return dict(zip(pkname, pkey))



"""A collection of utility methods used by various parts of the bidon package."""
import json
import re
import sys
from contextlib import contextmanager
from datetime import datetime, timezone
from uuid import UUID

from . import convert


_JSON_DEFAULTS = []


# pylint: disable=invalid-name
try_parse_int = convert.try_wrapper(convert.to_int)
try_parse_float = convert.try_wrapper(convert.to_float)
try_parse_decimal = convert.try_wrapper(convert.to_decimal)
try_parse_date = convert.try_wrapper(convert.to_date)
try_parse_datetime = convert.try_wrapper(convert.to_datetime)
# pylint: enable=invalid-name


def local_now():
  """Returns a datetime object for the current time, timezone unaware."""
  return datetime.now()


def utc_now():
  """Returns a datetime object for the current time, with the tzinfo set to timezone.utc"""
  return datetime.utcnow().replace(tzinfo=timezone.utc)


IS_UUID_RE = re.compile(r"^[a-z0-9]{8}(-[a-z0-9]{4}){3}-[a-z0-9]{12}$")
def is_uuid(val):
  """Checks whether the value is either an instance of UUID, or if it matches a uuid regex."""
  return val and (isinstance(val, UUID) or IS_UUID_RE.match(val) != None)


def exclude(source, keys, *, transform=None):
  """Returns a dictionary excluding keys from a source dictionary.

  keys can be either a set of keys, or a predicate function that accepts a key and returns either
  True or False indicating if it should be included.
  """
  check = keys if callable(keys) else lambda key: key in keys
  return {key: transform(source[key]) if transform else source[key] \
          for key in source if not check(key)}


def pick(source, keys, *, transform=None):
  """Returns a dictionary including only specified keys from a source dictionary.

  keys can be either a set of keys, or a predicate function that accepts a key and returns either
  True or False indicating if it should be included.
  """
  check = keys if callable(keys) else lambda key: key in keys
  return {key: transform(source[key]) if transform else source[key] \
          for key in source if check(key)}


def register_json_default(pred, convert_):
  """Register a predicate and converter function for json defaults.

  The predicate function is a function that must return true for the convert function to be called
  on an object when to_json is called.

  The registered functions are searched in order they are registered.
  """
  _JSON_DEFAULTS.append((pred, convert_))


def json_default(obj):
  """Convert an object to JSON, via the defaults set with register_json_default."""
  for default in _JSON_DEFAULTS:
    if default[0](obj):
      return default[1](obj)
  raise TypeError(repr(obj) + " is not JSON serializable")


def to_json(obj, pretty=False):
  """Converts an object to JSON, using the defaults specified in register_json_default."""
  sort_keys = False
  indent = None
  separators = (",", ":")

  if pretty:
    sort_keys = True
    indent = 2
    separators = (", ", ": ")

  return json.dumps(obj, sort_keys=sort_keys, indent=indent, separators=separators,
                    default=json_default)


def has_value(obj, name):
  """A flexible method for getting values from objects by name.

  returns:
  - obj is None: (False, None)
  - obj is dict: (name in obj, obj.get(name))
  - obj hasattr(name): (True, getattr(obj, name))
  - else: (False, None)
  """
  if obj == None:
    return (False, None)
  elif isinstance(obj, dict):
    return (name in obj, obj.get(name))
  elif hasattr(obj, name):
    return (True, getattr(obj, name))
  elif hasattr(obj, "__getitem__") and hasattr(obj, "__contains__") and name in obj:
    return (True, obj[name])
  else:
    return (False, None)


def get_value(obj, name, fallback=None):
  """Calls through to has_value. If has_value[0] is True, return has_value[1] otherwise returns
  fallback() if fallback is callable, else just fallback.
  """
  present, value = has_value(obj, name)
  if present:
    return value
  else:
    if callable(fallback):
      return fallback()
    else:
      return fallback


def set_value(obj, name, value):
  """A flexible method for setting a value on an object.

  If the object implements __setitem__ (such as a dict) performs obj[name] = value, else performs
  setattr(obj, name, value).
  """
  if hasattr(obj, "__setitem__"):
    obj[name] = value
  else:
    setattr(obj, name, value)


def with_defaults(method, nparams, defaults=None):
  """Call method with nparams positional parameters, all non-specified defaults are passed None"""
  args = [None] * nparams if not defaults else defaults + max(nparams - len(defaults), 0) * [None]
  return method(*args)


def namedtuple_with_defaults(ntup, defaults=None):
  """Wraps with_defaults for a named tuple."""
  return with_defaults(ntup, len(ntup._fields), defaults)


def delegate(from_owner, to_owner, methods):
  """Creates methods on from_owner to call through to methods on to_owner."""
  for method in methods:
    _delegate(from_owner, to_owner, method)


def _delegate(from_owner, to_owner, method):
  """Creates a method on from_owner to calls through to the same method on to_owner."""
  dgate = lambda self, *args, **kwargs: getattr(getattr(self, to_owner), method)(*args, **kwargs)
  dgate.__name__ = method
  dgate.__doc__ = "Delegates to {0}.{1}: {2}".format(to_owner, method, method.__doc__)
  setattr(from_owner, method, dgate)


def flatten_dict(source, ancestors=None):
  """Flattens a dictionary into (key, value) tuples. Where key is a tuple of ancestor keys."""
  if not ancestors:
    ancestors = ()
  for key in source:
    if isinstance(source[key], dict):
      yield from flatten_dict(source[key], ancestors + (key, ))
    else:
      yield (ancestors + (key, ), source[key])


@contextmanager
def get_file_object(filename, mode="r"):
  """Context manager for a file object. If filename is present, this is the
  same as with open(filename, mode): ...
  If filename is not present, then the file object returned is either
  sys.stdin or sys.stdout depending on the mode.
  """
  if filename is None:
    if mode.startswith("r"):
      yield sys.stdin
    else:
      yield sys.stdout
  else:
    with open(filename, mode) as fobj:
      yield fobj


from .field_mapping import FieldMapping
from . import terminal
from . import xml



"""Contains classes allowing the user to read data from various spreadsheet formats in a unified
manner.
"""

from .base import CellMode, Cell, WorksheetBase, WorkbookBase
from .csv import CSVWorkbook
from .excel import ExcelWorkbook
from .open_document import OpenDocumentWorkbook



import unittest
from uuid import uuid4

from bidon.data import transaction, autocommit, DataAccess

from tests import get_data_access, TEST_CALLPROC, TEST_ROWCOUNT


def rand():
  return str(uuid4())


class DataDataAccessTestCase(unittest.TestCase):
  @classmethod
  def setUpClass(cls):
    cls.dab = get_data_access()
    cls.dab.open()

  @classmethod
  def tearDownClass(cls):
    cls.dab.rollback()
    cls.dab.close()

  def test_execute(self):
    cr = self.dab.execute("select * from people where first_name = 'Yours'")
    r = cr.fetchone()
    self.assertEqual(r.last_name, "Truly")
    if TEST_ROWCOUNT:
      self.assertNotEqual(self.dab.row_count, -1)

  def test_callproc(self):
    if not TEST_CALLPROC: return
    cr, params = self.dab.callproc("test_proc1", (11, "yours"))
    rec = cr.fetchone()
    self.assertEqual(rec.first_name_p, "Yours")
    self.assertEqual(rec.last_name_p, "Truly")
    cr, params = self.dab.callproc("test_proc3", (11, 4))
    v = self.dab.get_scalar(cr)
    self.assertEqual(v, True)
    cr, params = self.dab.callproc("test_proc3", (11, 99))
    v = self.dab.get_scalar(cr)
    self.assertEqual(v, False)

  def test_find(self):
    rec = self.dab.find("people", dict(first_name="Yours"))
    self.assertEqual(rec.last_name, "Truly")
    rec = self.dab.find("people", dict(first_name="Ahab"))
    self.assertIsNone(rec)
    rec = self.dab.find("people", "first_name='Yours'")
    self.assertEqual(rec.last_name, "Truly")
    rec = self.dab.find("people", [("first_name", "Yours")])
    self.assertEqual(rec.last_name, "Truly")
    rec = self.dab.find("people", ("first_name = {0}".format(self.dab.core.sql_writer.to_placeholder(None)), ("Yours", )))
    self.assertEqual(rec.last_name, "Truly")

  def test_find_all(self):
    peeps = list(self.dab.find_all("people", dict(last_name="Michaels"), columns="first_name", order_by="first_name"))
    fns = [i.first_name for i in peeps]
    self.assertEqual(fns, ["Anthony", "Diamond"])
    peeps = list(self.dab.find_all("people", "id >= 10"))
    self.assertEqual(2, len(peeps))
    peeps = list(self.dab.find_all("people", [("id", 11)]))
    self.assertEqual(peeps[0].last_name, "Truly")
    peeps = list(self.dab.find_all("people", [("id", 10, ">=")]))
    self.assertEqual(2, len(peeps))
    peeps = list(self.dab.find_all("people", [("last_name", ("Truly", "Michaels"), "in")]))
    self.assertEqual(3, len(peeps))
    peeps = list(self.dab.find_all("people", [("id", (1, 5), "between")]))
    self.assertEqual(5, len(peeps))
    peeps = list(self.dab.find_all("people", order_by="id", limiting=(5, 0)))
    self.assertEqual(5, len(peeps))
    self.assertEqual([1, 2, 3, 4, 5], [p.id for p in peeps])
    peeps = list(self.dab.find_all("people", order_by="id", limiting=(5, 5)))
    self.assertEqual([6, 7, 8, 9, 10], [p.id for p in peeps])

  def test_page(self):
    peeps, count = self.dab.page("people", (0, 4), order_by="id")
    self.assertEqual(count, 11)
    self.assertEqual([1, 2, 3, 4], [p.id for p in peeps])
    peeps, count = self.dab.page("people", (1, 4), order_by="id")
    self.assertEqual(count, 11)
    self.assertEqual([5, 6, 7, 8], [p.id for p in peeps])

  def test_update(self):
    pid = 11
    da = self.dab

    def fp(id=pid):
      return da.find("people", dict(id=id))

    def trc(n=1):
      if TEST_ROWCOUNT:
        self.assertEqual(da.row_count, n)

    da.update("people", dict(last_name="Smith"), dict(id=pid))
    self.assertEqual(fp().last_name, "Smith")
    da.rollback()

    da.update("people", dict(last_name="Smith"), "id={0}".format(pid))
    trc()
    self.assertEqual(fp().last_name, "Smith")
    da.rollback()

    da.update("people", "last_name='Smith'", dict(id=pid))
    trc()
    self.assertEqual(fp().last_name, "Smith")
    da.rollback()

    da.update("people", "last_name='Smith'", "id={0}".format(pid))
    trc()
    self.assertEqual(fp().last_name, "Smith")
    da.rollback()

    da.update("people", "last_name='TrulyTruly'", [("id", (5, 6, 7, 11), "in")])
    trc(4)
    self.assertEqual(fp().last_name, "TrulyTruly")
    da.rollback()

    da.update("people", [("last_name", "Smith")], "id={0}".format(pid))
    trc()
    self.assertEqual(fp().last_name, "Smith")
    da.rollback()

    da.update("people", "last_name='Smith', first_name='Wotcha'", [("id", pid)])
    trc()
    rec = fp()
    self.assertEqual((rec.last_name, rec.first_name), ("Smith", "Wotcha"))
    da.rollback()

    da.update("people", [("last_name", "Smith")], [("id", 5, " <= ")])
    trc(5)
    rec = fp(1)
    self.assertEqual((rec.last_name, rec.id), ("Smith", 1))
    da.rollback()

    tc = dict(id=1)
    da.update("things", dict(other="something"), tc)
    self.assertEqual(da.find("things", tc).other, "something")
    da.update("things", dict(other=None), tc)
    self.assertIsNone(da.find("things", tc).other)
    da.update("things", dict(other="something"), tc)
    da.update("things", [("other", None)], [("id", tc["id"])])
    self.assertIsNone(da.find("things", tc).other)
    da.rollback()

    self.assertRaises(Exception, lambda: self.update("people", dict(last_name="Smith"), [("id", 5, "<=")]))

  def test_count(self):
    self.assertEqual(self.dab.count("people"), 11)

  def test_insert(self):
    da = self.dab
    da.insert("people", dict(first_name="A", last_name="B", age=55))
    self.assertEqual(da.count("people"), 12)
    da.rollback()

    da.insert("people", (("first_name", "A"), ("last_name", "B"), ("Age", 83)))
    self.assertEqual(da.count("people"), 12)
    da.rollback()

  def test_delete(self):
    da = self.dab
    da.delete("people", dict(id=11))
    self.assertEqual(da.count("people"), 10)
    da.rollback()

  def test_get_scalar(self):
    cr = self.dab.find_all("people", dict(id=11), columns="first_name")
    first_name = self.dab.get_scalar(cr)
    self.assertEqual(first_name, "Yours")

  def test_autocommit(self):
    da1 = get_data_access().open()
    da2 = get_data_access().open()

    da1.autocommit = True
    cnst1 = dict(first_name=rand(), last_name=rand(), age=1)
    da1.insert("people", cnst1)
    self.assertEqual(da2.find("people", cnst1).first_name, cnst1["first_name"])

    da1.autocommit = False
    cnst2 = dict(first_name=rand(), last_name=rand(), age=2)
    da1.insert("people", cnst2)
    self.assertIsNotNone(da1.find("people", cnst2))
    self.assertIsNone(da2.find("people", cnst2))

    da1.rollback()
    da1.delete("people", cnst1)
    da1.commit()
    self.assertEqual(da1.count("people"), 11)

  def test_transaction_contextmanager(self):
    da1 = get_data_access().open()
    da2 = get_data_access().open()
    da1.autocommit = False
    cnst = dict(first_name=rand(), last_name=rand(), age=3)
    with transaction(da1) as da:
      self.assertIsInstance(da, DataAccess)
    with transaction(da1):
      da1.insert("people", cnst)
    self.assertEqual(da2.find("people", cnst).first_name, cnst["first_name"])
    with transaction(da1):
      da1.delete("people", cnst)
    da1.autocommit = True
    self.assertEqual(da1.count("people"), 11)
    try:
      with transaction(da1):
        da1.insert("people", cnst)
        raise Exception()
    except Exception as ex:
      pass
    finally:
      self.assertIsNone(da1.find("people", cnst))

  def test_autocommit_contextmanager(self):
    da = get_data_access().open(False)
    cnst = dict(first_name=rand(), last_name=rand(), age=3)
    with transaction(da):
      with autocommit(da):
        da.insert("people", cnst)
        # Rollback should have no effect
        da.rollback()
    self.assertEqual(da.find("people", cnst).first_name, cnst["first_name"])
    da.delete("people", cnst)
    da.commit()



"""Classes and functions for interacting with DB API 2 databases."""
from .validation import Validation, Validator
from .model_base import ModelBase
from .sql_writer import SQLWriter
from .data_access_core import DataAccessCoreBase, InjectedDataAccessCore
from .data_access import DataAccess, transaction, autocommit
from .model_access import ModelAccess
from .foreign_model_wrapper import ForeignModelWrapper



"""Contains classes and methods for writing data to a terminal."""
import sys
from datetime import datetime


class ProgressPrinter(object):
  """A flexible command-line progress printer."""
  # pylint: disable=invalid-name

  def __init__(self, total, *, format="{} / {}", format_time_remaining=None):
    self._total = total
    self._format = format
    self._format_time_remaining = format_time_remaining
    self._current = 0
    self._started_at = None
    self._message = ""

  def __enter__(self):
    """Start the progress printer."""
    self.start()
    return self

  def __exit__(self, ex_type, ex_value, ex_traceback):
    """Stop the progress printer."""
    self.done()

  def start(self):
    """Sets several instance values to indicate when progress started."""
    self._started_at = datetime.now()
    self._current = 0

  def increment(self, amount=1, output=True):
    """Increment progress by a given amount (defaults to 1) and (optionally) output the progress
    status.
    """
    self._current += amount
    if output:
      # NOTE: This algorithm averages out the time it has taken to get to the current / total to
      #       determine how much time is left. So this is best used with jobs that have steps whose
      #       run time is roughly the same.

      now = datetime.now()
      elapsed_time = now - self._started_at
      expected_finish = self._started_at + elapsed_time * (1 / (self._current / self._total))
      remaining_time = expected_finish - now

      if self._format_time_remaining:
        remaining_time_str = self._format_time_remaining(remaining_time)
      else:
        remaining_time_str = str(remaining_time)

      tab_count = self._message.count("\t")
      sys.stdout.write("\r" + "\t" * tab_count + " " * (len(self._message) - tab_count))

      if isinstance(self._format, str):
        self._message = self._format.format(self._current, self._total, remaining_time_str)
      else:
        self._message = self._format(self._current, self._total, remaining_time_str)

      sys.stdout.write("\r" + self._message)
      sys.stdout.flush()

  def done(self):
    """Clean up after progress printing is done."""
    print()

  @staticmethod
  def ratio(current, total, time_remaining):
    """Returns the progress ratio."""
    return "{} / {}".format(current, total)

  @staticmethod
  def percentage(current, total, time_remaining):
    """Returns the progress percentage."""
    return "{}% completed".format(int(current / total * 100))

  @staticmethod
  def time_remaining(current, total, time_remaining):
    """Returns the time remaining."""
    return "~{} remaining".format(time_remaining)

  @staticmethod
  def ratio_and_percentage(current, total, time_remaining):
    """Returns the progress ratio and percentage."""
    return "{} / {} ({}% completed)".format(current, total, int(current / total * 100))

  @staticmethod
  def ratio_with_time_remaining(current, total, time_remaining):
    """Returns the progress ratio and time remaining."""
    return "{} / {} (~{} remaining)".format(current, total, time_remaining)

  @staticmethod
  def percentage_with_time_remaining(current, total, time_remaining):
    """Returns the progress percentage and time remaining."""
    return "{}% (~{} remaining)".format(int(current / total * 100), time_remaining)

  @staticmethod
  def ratio_and_percentage_with_time_remaining(current, total, time_remaining):
    """Returns the progress ratio, percentage and time remaining."""
    return "{} / {} ({}% completed) (~{} remaining)".format(
      current,
      total,
      int(current / total * 100),
      time_remaining)

  @staticmethod
  def time_remaining_whole_seconds(time_remaining):
    """Strips off the microsecond portion of time remaining."""
    return str(time_remaining).split(".")[0]


def table_to_string(headers, table, align="", *, lines=("-", "-+-", " | ")):
  """Write a list of headers and a table of rows to the terminal in a nice format.

  Parameters:
  - headers: a list of strings that are the headers of the table
  - table: a list of lists, the actual data to be printed.
  - align: a string whose elements are in the set {'<', '^', '>'}. These define how the values
           printed in the cells will align. '<': left, '^': center, '>': right. This argument is
           optional and all unspecified columns will align as if '<' were passed for them.
  - lines: a tuple of line characters to print for the table: (row_sep, row_intersection, col_sep)
  """
  header_separator, header_junction, row_separator = lines
  align = ("{0:<<" + str(len(headers)) + "}").format(align or "")
  all_lens = [tuple(len(c) for c in r) for r in table]

  if headers:
    all_lens.append(tuple(len(h) for h in headers))

  max_lens = [max(r[i] for r in all_lens) for i in range(len(headers))]
  col_outs = ["{{{0}: {1}{2}}}".format(i, align[i], w) for i, w in enumerate(max_lens)]
  fmt_str = row_separator.join(col_outs)
  if headers:
    yield fmt_str.format(*headers)
    yield header_junction.join((header_separator * ml for ml in max_lens))
  for row in table:
    yield fmt_str.format(*row)



"""Contains an implementation of WorkbookBase and WorksheetBase for working with .xls and .xlsx
documents, usig the xlrd library.
"""
try:
  import xlrd
except ImportError:
  pass

from .base import CellMode, WorksheetBase, WorkbookBase


class ExcelWorksheet(WorksheetBase):
  """Specialization of WorksheetBase for working with excel documents."""
  def __init__(self, raw_sheet, ordinal, handle_ambiguous_date):
    """Initialize the ExcelWorksheet instance."""
    super().__init__(raw_sheet, ordinal)
    self.name = self.raw_sheet.name
    self.nrows = self.raw_sheet.nrows
    self.ncols = self.raw_sheet.ncols
    self.handle_ambiguous_date = handle_ambiguous_date

  def parse_cell(self, cell, coords, cell_mode=CellMode.cooked):
    """Parses a cell according to the cell.ctype."""
    # pylint: disable=too-many-return-statements
    if cell_mode == CellMode.cooked:
      if cell.ctype == xlrd.XL_CELL_BLANK:
        return None

      if cell.ctype == xlrd.XL_CELL_BOOLEAN:
        return cell.value

      if cell.ctype == xlrd.XL_CELL_DATE:
        if self.handle_ambiguous_date:
          try:
            return self._parse_date(cell.value)
          except xlrd.xldate.XLDateAmbiguous:
            return self.handle_ambiguous_date(cell.value)
        else:
          return self._parse_date(cell.value)

      if cell.ctype == xlrd.XL_CELL_EMPTY:
        return None

      if cell.ctype == xlrd.XL_CELL_ERROR:
        return cell.value

      if cell.ctype == xlrd.XL_CELL_NUMBER:
        return cell.value

      if cell.ctype == xlrd.XL_CELL_TEXT:
        return cell.value

      raise ValueError("Unhandled cell type {0}".format(cell.ctype))
    else:
      return cell

  def get_row(self, row_index):
    """Returns the row at row_index."""
    return self.raw_sheet.row(row_index)

  def merged_cell_ranges(self):
    """Generates the sequence of merged cell ranges in the format:

    ((col_low, row_low), (col_hi, row_hi))
    """
    for rlo, rhi, clo, chi in self.raw_sheet.merged_cells:
      yield ((clo, rlo), (chi, rhi))

  def get_note(self, coords):
    """Get the note for the cell at the given coordinates.

    coords is a tuple of (col, row)
    """
    col, row = coords
    note = self.raw_sheet.cell_note_map.get((row, col))
    return note.text if note else None

  def _parse_date(self, cell_value):
    """Attempts to parse a cell_value as a date."""
    date_tuple = xlrd.xldate_as_tuple(cell_value, self.raw_sheet.book.datemode)
    return self.tuple_to_datetime(date_tuple)


class ExcelWorkbook(WorkbookBase):
  """Specialization of WorkbookBase for working with Excel workbooks."""
  def __init__(self, filename, formatting_info=False, handle_ambiguous_date=None):
    """Initialize the ExcelWorkbook instance."""
    super().__init__(filename)
    self.workbook = xlrd.open_workbook(self.filename, formatting_info=formatting_info)
    self.handle_ambiguous_date = handle_ambiguous_date

  def iterate_sheets(self):
    """Generates the sequence of sheets in the workbook."""
    for sheet in self.workbook.sheets():
      yield sheet

  def get_worksheet(self, raw_sheet, index):
    """Creates an ExcelWorksheet instance from a raw sheet."""
    return ExcelWorksheet(raw_sheet, index, self.handle_ambiguous_date)



import re
import unittest
import xml.etree.ElementTree as ET

from bidon.data import ModelBase, ForeignModelWrapper


class Person(ForeignModelWrapper):
  attrs = dict(
    first_name=None,
    last_name=None,
    company_id=None)
  mapping = dict(
    first_name=("first_name", lambda s: s.strip()),
    last_name=("last_name", lambda s: s.strip()),
    company_id=("company", lambda d: int(d["id"])))


class Company(ModelBase):
  attrs = dict(id=None, name=None)


class PersonXML(ForeignModelWrapper):
  attrs = dict(id=None, first_name=None, last_name=None, company=None)
  mapping = dict(
    id=("id", lambda i: int(i), None),
    first_name=("first_name", ),
    last_name=("last_name", ),
    company=("company", dict(
      id=("id", int, None),
      name=("name", ))))
  submodels=dict(company=Company)


class Simple(ForeignModelWrapper):
  attrs = dict(id=None, name=None)
  mapping = dict(id=("id", lambda d: int(d)), name=("name", lambda i: i))


class DataForeignModelWrapperTestCase(unittest.TestCase):
  def test_create_raw(self):
    d = {
      "first_name": "Trey",
      "last_name": "Cucco",
      "company": {
        "id": 188,
        "name": "ACME" }}
    p = Person.create(d)
    self.assertEqual(p.first_name, "Trey")
    self.assertEqual(p.last_name, "Cucco")
    self.assertEqual(p.company_id, 188)

  def test_map(self):
    l = [dict(id=1, name="Trey Cucco"),
         dict(id=2, name="J.R.R. Tolkien"),
         dict(id=3, name="C.S. Lewis")]
    ml = list(Simple.map(l))

    self.assertEqual(len(ml), 3)
    for s in ml:
      self.assertIsInstance(s, Simple)
    self.assertEqual(ml[0].id, 1)
    self.assertEqual(ml[0].name, "Trey Cucco")
    self.assertEqual(ml[1].id, 2)
    self.assertEqual(ml[1].name, "J.R.R. Tolkien")
    self.assertEqual(ml[2].id, 3)
    self.assertEqual(ml[2].name, "C.S. Lewis")

  def test_create_xml(self):
    xmlt = """<person id="1">
      <first_name>Trey</first_name>
      <last_name>Cucco</last_name>
      <company id="1">
        <name>ACME</name>
      </company>
    </person>"""
    root = ET.fromstring(xmlt)
    p = PersonXML.create(root, mapping_format="xml")
    self.assertEqual(p.id, 1)
    self.assertEqual(p.company.id, 1)
    self.assertEqual(p.company.name, "ACME")
    self.assertEqual(p.first_name, "Trey")
    self.assertEqual(p.last_name, "Cucco")
